import random
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import scipy.stats as stats
from jinja2 import Template
from IPython.display import HTML
import base64
from io import BytesIO
import pandas as pd

from wordcloud import WordCloud


from ..datahub import DataHub
from ..dtype import *

from .tables import *


"""
This module implements the visualization for the plot(df) function.
"""  


        
def hist_plot_thumb_mini(data,cfg, num_bins=None):

    # Exclude NaN values
    data_non_nan = data.dropna()
    
    unique = data.nunique()
    fig_histogram, ax_histogram = plt.subplots(figsize=(3, 0.4))

    if num_bins is None:
        num_bins = min(unique, 30)  # Adjust 40 to your preference or remove for automatic calculation

    ax_histogram.hist(data_non_nan, bins=num_bins, color=cfg['mini_hist_color'], edgecolor='white') #color='#06c'
    ax_histogram.yaxis.set_visible(False)
    ax_histogram.set_yscale('log')  # Set logarithmic scale on the y-axis TO HANDLE VISUALIZE OUTLIERS!!!
    ax_histogram.xaxis.set_visible(False)
    ax_histogram.set_frame_on(False)
    plt.grid(False)

    # Save each figure to a buffer
    buffer_histogram = BytesIO()
    fig_histogram.savefig(buffer_histogram, format='png')
    image_histogram_thumb_mini = base64.b64encode(buffer_histogram.getvalue()).decode('utf-8')
    plt.close(fig_histogram)

    return image_histogram_thumb_mini
    
def hist_plot(data,cfg):
        fig_histogram, ax_histogram = plt.subplots(figsize=(6, 4))
        ax_histogram.hist(data, bins=30, color='skyblue', edgecolor='black')
        ax_histogram.set_title('Histogram')
        ax_histogram.set_xlabel('Value')
        ax_histogram.set_ylabel('Frequency')
    
        # Save each figure to a buffer
        buffer_histogram = BytesIO()
        fig_histogram.savefig(buffer_histogram, format='png')
        image_histogram = base64.b64encode(buffer_histogram.getvalue()).decode('utf-8')
        plt.close(fig_histogram)

        return image_histogram
'''testing hist + kde '''

# def hist_plot_thumb(data,cfg, name=None):
#     # Create a histogram
#     fig_histogram, ax_histogram = plt.subplots(figsize=(6, 4))
#     ax_histogram.hist(data, bins=30, color=cfg['hist_color'], edgecolor='white', density=True) #'skyblue'
    
#     # Add a kernel density plot
#     sns.kdeplot(data, color=cfg['hist_kde_color'], ax=ax_histogram, warn_singular=False) #'pink'
    
#     ax_histogram.tick_params(axis='both', colors='#a9a9a9')
#     ax_histogram.set_xlabel('Value', color='#a9a9a9')
#     ax_histogram.set_ylabel('Frequency', color='#a9a9a9')
#     if name is not None:
#         ax_histogram.set_title(name)
#     ax_histogram.set_frame_on(False)
#     plt.grid(False)

#     # Save the figure to a buffer
#     buffer_histogram = BytesIO()
#     fig_histogram.savefig(buffer_histogram, format='png')
#     image_histogram_thumb = base64.b64encode(buffer_histogram.getvalue()).decode('utf-8')
#     plt.close(fig_histogram)

#     return image_histogram_thumb

    
def kde_plot(data):
        fig_kde, ax_kde = plt.subplots(figsize=(6, 4))
        sns.kdeplot(data, color='salmon')
        ax_kde.set_title('KDE Plot')
        ax_kde.set_xlabel('Value')
        ax_kde.set_ylabel('Density')

        # Save figure to a buffer
        buffer_kde = BytesIO()
        fig_kde.savefig(buffer_kde, format='png')
        image_kde = base64.b64encode(buffer_kde.getvalue()).decode('utf-8')
        plt.close(fig_kde)

        return image_kde
    
def qq_plot(data,cfg):
        # Exclude NaN values
        data_non_nan = data.dropna()
    
        fig_qq, ax_qq = plt.subplots(figsize=(6, 4))
        stats.probplot(data_non_nan, plot=ax_qq, fit=True, rvalue=True)
        ax_qq.set_title('Normal Q-Q Plot')
        ax_qq.set_xlabel('Theoretical Quantiles')
        ax_qq.set_ylabel('Ordered Values')

        # Save figure to a buffer
        buffer_qq = BytesIO()
        fig_qq.savefig(buffer_qq, format='png')
        image_qq = base64.b64encode(buffer_qq.getvalue()).decode('utf-8')
        plt.close(fig_qq)

        return image_qq


def box_plot(data):
        fig_box, ax_box = plt.subplots(figsize=(6, 4))
        sns.boxplot(data, color='lightgreen')
        ax_box.set_title('Box Plot')
        ax_box.set_xlabel('Value')

        # Save figure to a buffer
        buffer_box = BytesIO()
        fig_box.savefig(buffer_box, format='png')
        image_box = base64.b64encode(buffer_box.getvalue()).decode('utf-8')
        plt.close(fig_box)

        return image_box

def bar_chart(data):
    """
    Generate a horizontal bar chart.

    Parameters:
    - data (pandas.Series or pandas.DataFrame): Data for the bar chart.

    Returns:
    - str: Base64-encoded PNG image of the bar chart.
    """
    if isinstance(data, pd.DataFrame):
        data = data.squeeze()  # Convert DataFrame to Series if needed

    value_counts = data.value_counts(normalize=True) * 100
    
    fig_bar, ax_bar = plt.subplots(figsize=(6, 4))
    value_counts.plot(kind='barh', color='skyblue', edgecolor='black', ax=ax_bar)
    
    ax_bar.set_title('Horizontal Bar Chart')
    ax_bar.set_xlabel('Count')
    ax_bar.set_ylabel('Category')

    # Set explicit ticks and add percentage signs
    ticks = ax_bar.get_xticks()
    tick_labels = [f'{tick:.2f}%' for tick in ticks]
    ax_bar.set_xticks(ticks)
    ax_bar.set_xticklabels(tick_labels)

    
    # Save figure to a buffer
    buffer_bar = BytesIO()
    fig_bar.savefig(buffer_bar, format='png')
    image_bar = base64.b64encode(buffer_bar.getvalue()).decode('utf-8')
    plt.close(fig_bar)

    return image_bar



"""BIVARIATE"""

def scatter_plot(x, y):
    fig_scatter, ax_scatter = plt.subplots(figsize=(6, 4))
    ax_scatter.scatter(x, y, color='coral', alpha=0.6)
    ax_scatter.set_title('Scatter Plot')
    ax_scatter.set_xlabel('X')
    ax_scatter.set_ylabel('Y')

    # Save figure to a buffer
    buffer_scatter = BytesIO()
    fig_scatter.savefig(buffer_scatter, format='png')
    image_scatter = base64.b64encode(buffer_scatter.getvalue()).decode('utf-8')
    plt.close(fig_scatter)

    return image_scatter

def hexbin_plot(x, y):
    fig_hexbin, ax_hexbin = plt.subplots(figsize=(6, 4))
    ax_hexbin.hexbin(x, y, gridsize=30, cmap='Blues')
    ax_hexbin.set_title('Hexbin Plot')
    ax_hexbin.set_xlabel('X')
    ax_hexbin.set_ylabel('Y')

    # Save figure to a buffer
    buffer_hexbin = BytesIO()
    fig_hexbin.savefig(buffer_hexbin, format='png')
    image_hexbin = base64.b64encode(buffer_hexbin.getvalue()).decode('utf-8')
    plt.close(fig_hexbin)

    return image_hexbin
    
""" doesnt work check"""

# def pair_plot(dataframe):
#     fig_pair, ax_pair = plt.subplots(figsize=(8, 8))
#     sns.pairplot(dataframe, palette='viridis', diag_kind='kde', markers='o')
#     ax_pair.set_title('Pair Plot')

#     # Save figure to a buffer
#     buffer_pair = BytesIO()
#     fig_pair.savefig(buffer_pair, format='png')
#     image_pair = base64.b64encode(buffer_pair.getvalue()).decode('utf-8')
#     plt.close(fig_pair)

#     return image_pair

# def correlation_heatmap(dataframe):
#     fig_heatmap, ax_heatmap = plt.subplots(figsize=(8, 6))
#     sns.heatmap(dataframe.corr(), annot=True, cmap='coolwarm', fmt=".2f")
#     ax_heatmap.set_title('Correlation Heatmap')

#     # Save figure to a buffer
#     buffer_heatmap = BytesIO()
#     fig_heatmap.savefig(buffer_heatmap, format='png')
#     image_heatmap = base64.b64encode(buffer_heatmap.getvalue()).decode('utf-8')
#     plt.close(fig_heatmap)

#     return image_heatmap

def box_plot_con(x, y):

    # Ensure x and y are numeric (convert if necessary)
    fig_box, ax_box = plt.subplots(figsize=(8, 6))
    sns.boxplot(x=x, y=y, ax=ax_box)
    
    ax_box.set_title('Box Plot: Numerical vs Categorical')
    ax_box.set_xlabel('Categorical Variable')
    ax_box.set_ylabel('Numerical Variable')

    # Save figure to a buffer
    buffer_box = BytesIO()
    fig_box.savefig(buffer_box, format='png')
    image_box_con = base64.b64encode(buffer_box.getvalue()).decode('utf-8')
    plt.close(fig_box)

    return image_box_con

def line_chart(x, y):
    fig_line, ax_line = plt.subplots(figsize=(6, 4))
    ax_line.plot(x, y, marker='o', color='purple', linestyle='-', linewidth=2, markersize=8)
    ax_line.set_title('Line Chart')
    ax_line.set_xlabel('X')
    ax_line.set_ylabel('Y')

    # Save figure to a buffer
    buffer_line = BytesIO()
    fig_line.savefig(buffer_line, format='png')
    image_line = base64.b64encode(buffer_line.getvalue()).decode('utf-8')
    plt.close(fig_line)

    return image_line


def heatmap_plot(x, y):
    # Create a DataFrame from the categorical variables
    df = pd.DataFrame({'X': x, 'Y': y})

    # Create a pivot table
    pivot_table = df.pivot_table(index='Y', columns='X', aggfunc=len, fill_value=0)

    # Create a heatmap
    plt.figure(figsize=(8, 6))
    sns.heatmap(pivot_table, annot=True, cmap='YlGnBu', fmt='g')
    plt.title('Heatmap Plot')
    plt.xlabel('X')
    plt.ylabel('Y')

    # Save figure to a buffer
    buffer_heatmap = BytesIO()
    plt.savefig(buffer_heatmap, format='png')
    image_heatmap = base64.b64encode(buffer_heatmap.getvalue()).decode('utf-8')
    plt.close()

    return image_heatmap

def stacked_bar_chart_percentage(x, y):
    # Create a DataFrame from the categorical variables
    df = pd.DataFrame({'X': x, 'Y': y})

    # Create a cross-tabulation
    cross_tab = pd.crosstab(df['X'], df['Y'])

    # Normalize the data to get percentages
    cross_tab_percentage = cross_tab.div(cross_tab.sum(axis=1), axis=0) * 100

    # Create a stacked bar chart with percentages
    plt.figure(figsize=(8, 6))
    cross_tab_percentage.plot(kind='bar', stacked=True, colormap='Set3')

    plt.title('Stacked Bar Chart (Percentage)')
    plt.xlabel('X')
    plt.ylabel('Percentage')

    # Save figure to a buffer
    buffer_stacked_bar_percentage = BytesIO()
    plt.savefig(buffer_stacked_bar_percentage, format='png')
    image_stacked_bar_percentage = base64.b64encode(buffer_stacked_bar_percentage.getvalue()).decode('utf-8')
    plt.close()

    return image_stacked_bar_percentage

def nested_bar_chart(x, y):
    # Create a DataFrame from the categorical variables
    df = pd.DataFrame({'X': x, 'Y': y})

    # Create a count plot
    plt.figure(figsize=(8, 6))
    sns.countplot(x='X', hue='Y', data=df, palette='Set3')

    plt.title('Nested Bar Chart')
    plt.xlabel('X')
    plt.ylabel('Count')
    plt.legend(title='Y')

    # Save figure to a buffer
    buffer_nested_bar = BytesIO()
    plt.savefig(buffer_nested_bar, format='png')
    image_nested_bar = base64.b64encode(buffer_nested_bar.getvalue()).decode('utf-8')
    plt.close()

    return image_nested_bar


def nested_bar_chart(x, y):
    # Create a DataFrame from the categorical variables
    df = pd.DataFrame({'X': x, 'Y': y})

    # Create a count plot
    plt.figure(figsize=(8, 6))
    sns.countplot(x='X', hue='Y', data=df, palette='Set3')

    plt.title('Nested Bar Chart')
    plt.xlabel('X')
    plt.ylabel('Count')
    plt.legend(title='Y')

    # Save figure to a buffer
    buffer_nested_bar = BytesIO()
    plt.savefig(buffer_nested_bar, format='png')
    image_nested_bar = base64.b64encode(buffer_nested_bar.getvalue()).decode('utf-8')
    plt.close()

    return image_nested_bar


###########################################################################################################################
######################################### TEXT CHARTS ###########################################################
###########################################################################################################################



def word_cloud(data, cfg, name=None):
    # Combine all text in the specified column
    text = ' '.join(data.astype(str))
    
    cfg = {
    'width': 300,
    'height': 200,
    'fig_width': 5,
    'fig_height': 3.334,
    'max_words': 50,
    }

    # Generate the word cloud
    wordcloud = WordCloud(width=cfg['width'], height=cfg['height'], background_color='white',max_words=cfg['max_words']).generate(text)

    # Access the tokenized words using the process_text method
    tokens = wordcloud.process_text(text)

    # Plot the WordCloud image
    fig_wordcloud, ax_wordcloud = plt.subplots(figsize=(cfg['fig_width'], cfg['fig_height']))
    ax_wordcloud.imshow(wordcloud, interpolation='bilinear')
    ax_wordcloud.axis('off')  # Turn off axis labels
    if name is not None:
        ax_wordcloud.set_title(name)
    ax_wordcloud.set_frame_on(False)
    plt.grid(False)

    # Save the figure to a buffer
    buffer_wordcloud = BytesIO()
    fig_wordcloud.savefig(buffer_wordcloud, format='png')
    image_wordcloud_thumb = base64.b64encode(buffer_wordcloud.getvalue()).decode('utf-8')
    plt.close(fig_wordcloud)

    return image_wordcloud_thumb, tokens

# Example usage
# Assuming you have a DataFrame named 'df' with a column 'text_column'






###########################################################################################################################
######################################### CATEGORICAL CHARTS ###########################################################
###########################################################################################################################


def bar_chart_thumb(data, cfg, threshold=0.1):
    """
    Generate a horizontal bar chart.

    Parameters:
    - data (pandas.Series or pandas.DataFrame): Data for the bar chart.
    - threshold (float): Proportion threshold to decide whether to place the count inside or outside the bar.

    Returns:
    - str: Base64-encoded PNG image of the bar chart.
    """
    if isinstance(data, pd.DataFrame):
        data = data.squeeze()  # Convert DataFrame to Series if needed

    value_counts = data.value_counts()

    # Check if there are more than 5 unique values
    if len(value_counts) > 5:
        # Combine smaller categories into 'others'
        others_count = value_counts.iloc[5:].sum()
        value_counts = pd.concat([value_counts.head(5), pd.Series([others_count], index=[' others'])])

    total_count = value_counts.sum()

    # Calculate a proportional width based on the number of bars
    max_width = 0.8
    width = max_width if len(value_counts) > 2 else 0.4

    fig_bar, ax_bar = plt.subplots(figsize=(8, 3))  # Customizable width and height
    bars = value_counts.plot(kind='barh', color=cfg['bar_color'], edgecolor='none', ax=ax_bar, width=width)  # Set edgecolor to 'none' for no edges

    ax_bar.xaxis.set_visible(False)

    ax_bar.set_frame_on(False)
    plt.grid(False)

    # Hide y-axis label
    ax_bar.set_ylabel('')

    # Increase the size of the y-axis ticks
    ax_bar.tick_params(axis='y', which='major', labelsize=12)  # Adjust the labelsize as needed

    # Add names at the center or outside of each bar with a larger font size
    for bar, name in zip(bars.patches, value_counts.index):
        width, height = bar.get_width(), bar.get_height()
        count = int(width)
        proportion = width / total_count

        if proportion < threshold:
            ax_bar.text(width + 0.01, bar.get_y() + height/2, f'{name}', va='center', ha='left', color='#636363', size=14)
        else:
            ax_bar.text(width/2, bar.get_y() + height/2, f'{name}', va='center', ha='center', color='#636363', size=14)

    # Set a maximum of 6 ticks
    num_ticks = min(len(value_counts), 6)
    
    # Set y-axis ticks and labels
    ax_bar.set_yticks(range(num_ticks))
    # Assuming your keys are the same as the range of the length of the array
    ax_bar.set_yticklabels(value_counts.values)


    ax_bar.yaxis.set_visible(True)
    #print(value_counts.values)
    # Save figure to a buffer
    buffer_bar = BytesIO()
    fig_bar.savefig(buffer_bar, format='png')
    image_bar_thumb = base64.b64encode(buffer_bar.getvalue()).decode('utf-8')
    plt.close(fig_bar)

    return image_bar_thumb


def pie_chart(data_series, cfg, name=None):
    # Use value_counts to get counts and labels
    value_counts = data_series.value_counts()

    # Extract data and labels from the value_counts result
    data = value_counts.values
    labels = value_counts.index

    # Limit to the top 5 categories
    top_categories = 5
    if len(labels) > top_categories:
        data = data[:top_categories]
        labels = labels[:top_categories]
        # Combine the remaining categories into "Others"
        data = np.append(data, sum(value_counts.values[top_categories:]))
        labels = np.append(labels, 'Others')

    # Create a pie chart
    fig_pie, ax_pie = plt.subplots(figsize=(6, 4))
    ax_pie.pie(data, labels=labels, autopct='%1.1f%%', colors=sns.color_palette('pastel'), startangle=90)

    ax_pie.tick_params(axis='both', colors='#a9a9a9')
    if name is not None:
        ax_pie.set_title(name)
    ax_pie.set_frame_on(False)
    plt.grid(False)

    # Save the figure to a buffer
    buffer_pie = BytesIO()
    fig_pie.savefig(buffer_pie, format='png')
    image_pie_thumb = base64.b64encode(buffer_pie.getvalue()).decode('utf-8')
    plt.close(fig_pie)

    return image_pie_thumb

# def Nightndale_chart(data_series, cfg, name=None):
#     #Use value_counts to get counts and labels
#     value_counts = data_series.value_counts()

#     # Extract data and labels from the value_counts result
#     data = value_counts.values
#     labels = value_counts.index

#     # Limit to the top 5 categories
#     top_categories = 10
#     if len(labels) > top_categories:
#         data = data[:top_categories]
#         labels = labels[:top_categories]
#         # Combine the remaining categories into "Others"
#         data = np.append(data, sum(value_counts.values[top_categories:]))
#         labels = np.append(labels, 'Others')

#     # Convert data to radians
#     theta = np.linspace(0.0, 2 * np.pi, len(labels), endpoint=False)

#     # Create a Nightingale diagram
#     fig, ax = plt.subplots(subplot_kw={'projection': 'polar'}, figsize=(6, 4))
#     ax.fill(theta, data, color='skyblue', alpha=0.75)

#     # Customize the plot
#     ax.set_theta_offset(np.pi / 2)  # Adjust the starting angle if needed
#     ax.set_theta_direction(-1)  # Change to 1 for clockwise direction
#     ax.set_rlabel_position(0)  # Move radial labels to the center
#     ax.set_yticklabels([])  # Hide radial labels if not needed

#     # Add category labels
#     ax.set_xticks(theta)
#     ax.set_xticklabels(labels)

#     if name is not None:
#         ax.set_title(name)

#     # Save the figure to a buffer
#     buffer_nightingale = BytesIO()
#     fig.savefig(buffer_nightingale, format='png')
#     image_nightingale_thumb = base64.b64encode(buffer_nightingale.getvalue()).decode('utf-8')
#     plt.close(fig)

#     return image_nightingale_thumb



# def radial_bar_chart(data_series,cfg):
#     # Extract up to 5 categories and values from the data series
#     value_counts = data_series.value_counts()

#     # Extract data and labels from the value_counts result
#     data = value_counts.values
#     labels = value_counts.index

#     # Limit to the top categories
#     top_categories = 10
#     if len(labels) > top_categories:
#         data = data[:top_categories]
#         labels = labels[:top_categories]
#         # Combine the remaining categories into "Others"
#         data = np.append(data, sum(value_counts.values[top_categories:]))
#         labels = np.append(labels, 'Others')

#     theta = np.linspace(0.0, 2 * np.pi, len(labels), endpoint=False)

#     fig, ax = plt.subplots(subplot_kw={'projection': 'polar'})
#     ax.bar(theta, data, color=['skyblue', 'orange', 'lightgreen', 'pink', 'lightcoral'], alpha=0.7)

#     ax.set_theta_zero_location('N')
#     ax.set_theta_direction(-1)
#     ax.set_rlabel_position(0)  # Move radial labels to the center
#     ax.set_yticklabels([])  # Hide radial labels if not needed

#     ax.set_xticks(theta)
#     ax.set_xticklabels(labels)

#     plt.title('Radial Bar Chart')

#     # Save figure to a buffer
#     buffer_radial = BytesIO()
#     plt.savefig(buffer_radial, format='png')
#     image_radial = base64.b64encode(buffer_radial.getvalue()).decode('utf-8')
#     plt.close(fig)

#     return image_radial




# def treemap(data_series, cfg):
#     # Use value_counts to get counts and labels
#     value_counts = data_series.value_counts()

#     # Extract data and labels from the value_counts result
#     data = value_counts.values
#     labels = value_counts.index
#     top_categories=5
#     # Limit to the top categories
#     if len(labels) > top_categories:
#         data = data[:top_categories]
#         labels = labels[:top_categories]
#         # Combine the remaining categories into "Others"
#         data = np.append(data, sum(value_counts.values[top_categories:]))
#         labels = np.append(labels, 'Others')

#     squarify.plot(sizes=data, label=labels, color=['skyblue', 'orange', 'lightgreen', 'pink', 'lightcoral'], alpha=0.7)
#     plt.axis('off')
#     plt.title('Treemap')

#     # Save figure to a buffer
#     buffer_treemap = BytesIO()
#     plt.savefig(buffer_treemap, format='png')
#     image_treemap = base64.b64encode(buffer_treemap.getvalue()).decode('utf-8')
#     plt.close()

#     return image_treemap



# def donut_chart(data_series, cfg):
#     # Use value_counts to get counts and labels
#     value_counts = data_series.value_counts()

#     # Extract data and labels from the value_counts result
#     data = value_counts.values
#     labels = value_counts.index
#     top_categories=5
#     # Limit to the top categories
#     if len(labels) > top_categories:
#         data = data[:top_categories]
#         labels = labels[:top_categories]
#         # Combine the remaining categories into "Others"
#         data = np.append(data, sum(value_counts.values[top_categories:]))
#         labels = np.append(labels, 'Others')

#     plt.pie(data, labels=labels, autopct='%1.1f%%', startangle=90, wedgeprops=dict(width=0.4), colors=['skyblue', 'orange', 'lightgreen', 'pink', 'lightcoral'])
#     plt.axis('equal')  # Equal aspect ratio ensures the donut is circular.
#     plt.title('Donut Chart')

#     # Save figure to a buffer
#     buffer_donut = BytesIO()
#     plt.savefig(buffer_donut, format='png')
#     image_donut = base64.b64encode(buffer_donut.getvalue()).decode('utf-8')
#     plt.close()

#     return image_donut

##################################################################################################################
def hist_plot_thumb(data,cfg, name=None):
    # Create a histogram
    fig_histogram, ax_histogram = plt.subplots(figsize=(6, 4))
    ax_histogram.hist(data, bins=30, color=cfg['hist_color'], edgecolor='white', density=True) #'skyblue'
    
    # Add a kernel density plot
    sns.kdeplot(data, color=cfg['hist_kde_color'], ax=ax_histogram, warn_singular=False) #'pink'
    
    ax_histogram.tick_params(axis='both', colors='#a9a9a9')
    ax_histogram.set_xlabel('Value', color='#a9a9a9')
    ax_histogram.set_ylabel('Frequency', color='#a9a9a9')
    if name is not None:
        ax_histogram.set_title(name)
    ax_histogram.set_frame_on(False)
    plt.grid(False)

    # Save the figure to a buffer
    buffer_histogram = BytesIO()
    fig_histogram.savefig(buffer_histogram, format='png')
    image_histogram_thumb = base64.b64encode(buffer_histogram.getvalue()).decode('utf-8')
    plt.close(fig_histogram)

    return image_histogram_thumb





def wiz_biv_cat(hub):
    """
    Create visualizations for plot(df, Continuous)
    """

    col1,col2, data = hub["col1"], hub["col2"], hub["data"]
    #print(col,data)
    #tabs: List[Panel] = []
    # pic = hist_plot(col)
    # pic2 = qq_plot(col)
    # pic3 = kde_plot(col)
    #htgs: Dict[str, List[Tuple[str, str]]] = {}






    store = {
        # 'val_tab2': value_tableH(data,pic),
        # 'val_tab':value_tableH(data,pic2,pic3),
        'hm': heatmap_plot(col1,col2),
        'nbch' : nested_bar_chart(col1,col2),
        'sbch'  : stacked_bar_chart_percentage(col1,col2),
        # 'box' : box_plot(col),
    }

    plots = {
            'type': ['a','b','c','d','e','f','g','h','i','j','k','l','m'
            ],
        
            'title': [#'iq',
                      #'Value Table', 
                      'HM', 
                      'Nest', 
                      'Stacked', 
                      #'Box Plot'
                    ],
        
            'image': [#None, 
                      #None, 
                      store['hm'], 
                      store['nbch'], 
                      store['sbch'],
                      #store['box']
                    ],
        
            'value_table': [#store['val_tab2'],
                            #store['val_tab'], 
                            None, 
                            None, 
                            None, 
                            #None
            ],
        }
    
    return plots



def wiz_biv_cat(hub):
    """
    Create visualizations for plot(df, Continuous)
    """

    col1,col2, data = hub["col1"], hub["col2"], hub["data"]
    #print(col,data)
    #tabs: List[Panel] = []
    # pic = hist_plot(col)
    # pic2 = qq_plot(col)
    # pic3 = kde_plot(col)
    #htgs: Dict[str, List[Tuple[str, str]]] = {}






    store = {
        # 'val_tab2': value_tableH(data,pic),
        # 'val_tab':value_tableH(data,pic2,pic3),
        'hm': heatmap_plot(col1,col2),
        'nbch' : nested_bar_chart(col1,col2),
        'sbch'  : stacked_bar_chart_percentage(col1,col2),
        # 'box' : box_plot(col),
    }

    plots = {
            'type': ['a','b','c','d','e','f','g','h','i','j','k','l','m'
            ],
        
            'title': [#'iq',
                      #'Value Table', 
                      'HM', 
                      'Nest', 
                      'Stacked', 
                      #'Box Plot'
                    ],
        
            'image': [#None, 
                      #None, 
                      (store['hm'],), 
                      (store['nbch'],), 
                      (store['sbch'],),
                      #store['box']
                    ],
        
            'value_table': [#store['val_tab2'],
                            #store['val_tab'], 
                            None, 
                            None, 
                            None, 
                            #None
            ],
        }
    
    return plots
def wiz_biv_con_cat(hub):
    """
    Create visualizations for plot(df, Continuous)
    """

    col1,col2, data = hub["col1"], hub["col2"], hub["data"]
    #print(col,data)
    #tabs: List[Panel] = []
    # pic = hist_plot(col)
    # pic2 = qq_plot(col)
    # pic3 = kde_plot(col)
    #htgs: Dict[str, List[Tuple[str, str]]] = {}






    store = {
        # 'val_tab2': value_tableH(data,pic),
        # 'val_tab':value_tableH(data,pic2,pic3),
        'bx': box_plot_con(col1,col2),
        'lc' : line_chart(col1,col2),
        # 'qq'  : qq_plot(col),
        # 'box' : box_plot(col),
    }

    plots = {
            'type': ['a','b','c','d','e','f','g','h','i','j','k','l','m'
            ],
        
            'title': [#'iq',
                      #'Value Table', 
                      'Box', 
                      'Line', 
                      #'QQ Plot', 
                      #'Box Plot'
                    ],
        
            'image': [#None, 
                      #None, 
                      (store['bx'],), 
                      (store['lc'],), 
                      #store['qq'],
                      #store['box']
                    ],
        
            'value_table': [#store['val_tab2'],
                            #store['val_tab'], 
                            None, 
                            None, 
                            #None, 
                            #None
            ],
        }
    
    return plots

def wiz_biv_con(hub):
    """
    Create visualizations for plot(df, Continuous)
    """

    col1,col2, data = hub["col1"], hub["col2"], hub["data"]
    #print(col,data)
    #tabs: List[Panel] = []
    # pic = hist_plot(col)
    # pic2 = qq_plot(col)
    # pic3 = kde_plot(col)
    #htgs: Dict[str, List[Tuple[str, str]]] = {}






    store = {
        # 'val_tab2': value_tableH(data,pic),
        # 'val_tab':value_tableH(data,pic2,pic3),
        'sc': scatter_plot(col1,col2),
        'hx' : hexbin_plot(col1,col2),
        # 'qq'  : qq_plot(col),
        # 'box' : box_plot(col),
    }

    plots = {
            'type': ['a','b','c','d','e','f','g','h','i','j','k','l','m' 
                     
            ],
        
            'title': [#'iq',
                      #'Value Table', 
                      'Scatter', 
                      'Hexbin', 
                      #'QQ Plot', 
                      #'Box Plot'
                    ],
        
            'image': [#None, 
                      #None, 
                      (store['sc'],), 
                      (store['hx'],), 
                      #store['qq'],
                      #store['box']
                    ],
        
            'value_table': [#store['val_tab2'],
                            #store['val_tab'], 
                            None, 
                            None, 
                            #None, 
                            #None
            ],
        }
    
    return plots

container_keysN = {
        'container1': ['VALUES', 'MISSING', 'DISTINCT', '', 'MEMORY', 'DTYPE'],
        'container2': ['MAX', '95%', 'Q3', 'AVG', 'MEDIAN', 'Q1', '5%', 'MIN'],
        'container3': ['RANGE', 'IQR', 'STD', 'VAR', ' ', 'KURT.', 'SKEW', 'SUM']
    }

container_keysT = {
        'container1': ['VALUES', 'MISSING', 'DISTINCT', '', 'MEMORY', 'DTYPE'],
        'container2': ['Max length', 'Mean length', 'Min length'],
        'container3': ['RANGE', 'IQR', 'STD', 'VAR', ' ', 'KURT.', 'SKEW', 'SUM']
    }

container_keysC = {
        'container1': ['VALUES', 'MISSING', 'DISTINCT', '', 'MEMORY', 'DTYPE'],
        'container3a': [0,1,2,3,4,5]
        #'container3': ['RANGE', 'IQR', 'STD', 'VAR', ' ', 'KURT.', 'SKEW', 'SUM']
    }


def wiz_text(hub,cfg):

    """
    Create visualizations for plot(df, Text)
    """

    col, data = hub["col"], hub["data"]

    w_len = data['w_len']
    #pic = hist_plot_thumb(w_len,cfg,name="Value Lengths")
    
    pic, tokens = word_cloud(col, cfg)

    


    from collections import Counter
    
    # Basic Analysis: Calculate word frequencies
    word_freq = Counter(tokens)
    most_common_words = word_freq.most_common(10)
    #print("Most common words:", most_common_words)
    ser_tokens = pd.Series(tokens).sort_values(ascending=False)
    #print(ser_tokens)
    store = {
        'stat_tab': generate_table(data,container_keysT,'val_tableT', style, pic,cfg=cfg),
        #'val_tab1': value_tableT(data,style,pic),
        'val_tabx':value_tableX(ser_tokens,cfg,tokens=True),
        
        #'hist': hist_plot(col),
        #'kde' : kde_plot(col),
        #'qq'  : qq_plot(col),
        #'box' : box_plot(col),
    }
    toto = tab_info(hub,cfg)

    plot_info = {
        'type': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm'],
        'title': [toto, 'Stats', 'Value Table'],
        'image': [None, None, None],
        'value_table': [None, store['stat_tab'], store['val_tabx']],
    }

    return plot_info


def wiz_cat(hub,cfg):
    """
    Create visualizations for plot(df, Categorical)
    """
    col, data = hub["col"], hub["data"]
    
    # Pass the v_type from comp already; here instead of categorical, it changes to categorical
    # v_type = 'categorical'
    # data['v_type'] = v_type

    pic = bar_chart_thumb(col,cfg)
    
    pie = pie_chart(col,cfg)
    #rad = radial_bar_chart(col,cfg)

    fake_dict = {1:'a',2:'b',3:'c',4:'d',5:'e'}
    data.update(fake_dict)
    
    toto = tab_info(hub,cfg)
    
    # Store the results in variables to avoid repeated dictionary lookups
    stat_tab = generate_table(data, container_keysC, 'val_tableC', style, pic,cfg=cfg)
    val_tabx = value_tableX(col,cfg)
    #hist = hist_plot(col,cfg)

    plots = {
        'type': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm'],
        'title': [toto, 'Stats', 'Pie Chart', 'Value Table'],
        'image': [None, None, (pie,), None],
        'value_table': [None, stat_tab, None, val_tabx],
    }
    
    return plots

def wiz_num(hub,cfg):
    """
    Create visualizations for plot(df, Continuous)
    """
    
    col, data = hub["col"], hub["data"]

    # im passing the v_type from comp already, here inseted of continuous, it changes is to numerical
    #v_type = 'numerical'
    #data['v_type'] = v_type
    
    pic_thumb = hist_plot_thumb(col,cfg)
    
    toto = tab_info(hub,cfg)
    
    # Store the results in variables to avoid repeated dictionary lookups
    stat_tab = generate_table(data, container_keysN, 'val_tableH', style, pic_thumb,cfg=cfg)
    val_tabx = value_tableX(col,cfg)
    qq = qq_plot(col,cfg)

    plots = {
        'type': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm'],
        'title': [toto, 'Stats', 'QQ Plot', 'Value Table'],
        'image': [None, None, (qq,), None],
        'value_table': [None, stat_tab, None, val_tabx],
    }
    
    return plots

def wiz_dt(hub,cfg):

    """
    Create visualizations for plot(df, dt)
    """
    col, data = hub["col"], hub["data"]
    
    # Pass the v_type from comp already; here instead of categorical, it changes to categorical
    # v_type = 'categorical'
    # data['v_type'] = v_type

    pic = hist_plot_thumb(col,cfg)
    #pic = bar_chart_thumb(col,cfg)
    #pie = pie_chart(col,cfg)
    
    toto = tab_info(hub,cfg)
    
    # Store the results in variables to avoid repeated dictionary lookups
    stat_tab = generate_table(data, container_keysN, 'val_tableH', style, pic,cfg=cfg)
    val_tabx = value_tableX(col,cfg)
    #hist = hist_plot(col,cfg)

    plots = {
        'type': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm'],
        'title': [toto, 'Stats', 'Value Table'],
        'image': [None, None, None],
        'value_table': [None, stat_tab, val_tabx],
    }
    
    return plots

def tab_info(hub,cfg):

    #print(hub)
    name = hub["col"].name    
    col, data = hub["col"], hub["data"]
    

    """" just copy of render num """
    pic_mini = hist_plot_thumb_mini(col,cfg)
    
    if data['MISSING_P'] > 0:
        font_size = '13px'
        font_color = '#FF6961'
        font_weight = '500'
    else:
        font_size = '11px'
        font_color = '#a9a9a9'
        font_weight = '300'
        
    if data['DISTINCT'] == 1 or data['DISTINCT_P'] == 100:
        font_sizeD = '13px'
        font_colorD = '#FF6961' #red
        font_weightD = '500'
    else:
        font_sizeD = '11px'
        font_colorD = '#a9a9a9'
        font_weightD = '300' 


    return f"""
        <div class="top_tab" style="width: 750px; height: 22px; padding: 0px 10px 0px; background-color: white; display: flex; align-items: center;">
            <div style="padding: 12px 10px 0px; width: 270px;">
                    <div style="font-size: 18px; height:20px;">{name}</div>
                    <div class="bottom_line" style="font-size: 12px; color: #a9a9a9; font-weight: 500; height:20px;">{data['v_type']}</div>
            </div>
                           <div class="table_tab" style="width: 180px; font-size: 10px; margin-bottom: 0px;">
                                <div style="display: flex;">
                        
                                    <div style="width: 60px; font-size: 11px; text-align: center; padding: 10px 5px; color: #a9a9a9; font-weight: 300;"> {data['VALUES_P']}%</div>
                                     <div style="width: 60px; font-size: {font_size}; text-align: center; padding: 10px 5px; color: {font_color}; font-weight: {font_weight};"> {data['MISSING_P']}%</div>
                                    <div style="width: 60px; font-size: {font_sizeD}; text-align: center; padding: 10px 5px; color: {font_colorD}; font-weight: {font_weightD};"> {data['DISTINCT_P']}%</div>
                                   
                                </div>
                                <div style="display: flex;">
                        
                                    
                                    <div style="width: 60px; font-size: 10px; text-align: center; padding: 0px 5px; color: #a9a9a9; font-weight: 300;"> VALID</div>
                                    <div style="width: 60px; font-size: 10px; text-align: center; padding: 0px 5px; color: #a9a9a9; font-weight: 300;"> MISSING</div>
                                    <div style="width: 60px; font-size: 10px; text-align: center; padding: 0px 5px; color: #a9a9a9; font-weight: 300;"> UNIQUE</div>
                                   
                                </div>
                            </div>
                       
            <div style="padding: 0px 10px 0px; width: 300px;">
                        <div class="table_tab_pic" style="width: 300px; height: 50px; font-size: 10px; margin-bottom: 0px;">
                            <div style="text-align: center; padding: 0px 5px; color: #a9a9a9; font-weight: 300; height: 40px;">
                                <img class="img-thumb-mini" src="data:image/png;base64, {pic_mini}" alt="Histogram" style="padding-top: 5px;">
                            </div>
                                <div style="display: flex;">
                                    <div style="width: 50px;"></div>
                                    <div style="width: 50px; font-size: 10px; text-align: left; padding: 0px 5px; color: #a9a9a9; font-weight: 300;"> Unique</div>
                                    <div style="width: 100px; font-size: 11px; text-align: left; padding: 0px 5px; color: {font_colorD}; font-weight: 500;"> {data['DISTINCT']}</div>
                                    <div style="width: 50px;"></div>
                                    <div style="width: 50px; font-size: 10px; text-align: left; padding: 0px 5px; color: #a9a9a9; font-weight: 300;"> Total</div>
                                    <div style="width: 100px; font-size: 11px; text-align: left; padding: 0px 5px; color: #a9a9a9; font-weight: 500;"> {data['TOTAL']}</div>
                                </div>
        
                        </div>
            </div>
        </div>
            """

def wiz(hub,cfg):
    """
    Render a basic plot
    Parameters
    ----------
    hub
        The DataHub containing results from the compute function.
    cfg
        Config instance
    """

    if hub.variable_type == "categorical":
        variable_data = wiz_cat(hub,cfg)
    elif hub.variable_type == "continuous":
        variable_data = wiz_num(hub,cfg)
    elif hub.variable_type == "text":
        variable_data = wiz_text(hub,cfg)
    elif hub.variable_type == "biv_con":
        variable_data = wiz_biv_con(hub)
    elif hub.variable_type == "biv_con_cat":
        variable_data = wiz_biv_con_cat(hub)
    elif hub.variable_type == "biv_cat":
        variable_data = wiz_biv_cat(hub)
    elif hub.variable_type == "datetime":
        variable_data = wiz_dt(hub, cfg)
    # elif hub.variable_type == "geography_column":
    #     variable_data = render_geo(hub, cfg)
    return variable_data


import pandas as pd
import hashlib

from ..wizrenderer import WizRenderer
from ..dtype import *
from ..configs import *

from .calc import calc
from .wiz import wiz
 


def eda(df, col1=None, col2=None, vars=None):
    # Configuration settings
    cfg = Configs.current_theme


    # Get dtypes
    dtypes = Configs.get_dtypes(df)

    
    # Check if caching is enabled
    use_cache = Configs.use_eda_cache

    # Assuming eda_cache is the dictionary you want to get the size of
    memory_size = sys.getsizeof(Configs.eda_cache)
    print(f"Memory size of eda_cache: {memory_size} bytes")

    # Generate a hash key for caching based on function parameters
    cache_key = hashlib.sha256(f"{col1}-{col2}-{dtypes}".encode()).hexdigest()

    # Check if result is cached
    if Configs.use_eda_cache and cache_key in Configs.eda_cache:
        cached_result = Configs.eda_cache[cache_key]
        cached_df = cached_result['df']

        # Now, compare the two DataFrames
        are_equal = df[col1].equals(cached_df[col1])
        if are_equal:
            # Return the cached result
            return WizRenderer(cached_result['data_load'], cached_result['variable_type'], Configs.current_theme)

    # Calculate
    hub = calc(df, col1, col2, cfg, dtypes, vars)
    if hub is None:  # because plot(df) is recursive
        return 0

    # Render
    data_load = wiz(hub, cfg)

    # Save the result to cache
    if use_cache:
        Configs.eda_cache[cache_key] = {
            'data_load': data_load,
            'variable_type': hub.variable_type,
            'df': df.copy()  # Cache a copy of the DataFrame
        }

    return WizRenderer(data_load, hub.variable_type, cfg)



import random
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import scipy.stats as stats
from jinja2 import Template
from IPython.display import HTML
import base64
from io import BytesIO
import pandas as pd


from ..datahub import DataHub
from ..dtype import *


style = """                
.my-custom-styles body {
    font-family: Arial, sans-serif;
}

.container {
    padding: 10px;
    margin: 1px;
    box-sizing: border-box;
}

.flex-container {
    display: flex;
    flex-wrap: wrap;  /* Add flex-wrap for responsive layout */
}

.flexing {
    flex-grow: 1;
}

.con-100 {
    width: 10px;
}

.container-200 {
    width: 200px;  /* Set the desired width directly */
}

.container-400 {
    width: 400px;  /* Set the desired width directly */
}

.container-500 {
    width: 488px;  /* Set the desired width directly */
}

.container-800 {
    width: 800px;  /* Set the desired width directly */
}

.img-container {
    width: 100%;
    text-align: center;
}

.img-thumb {
    overflow: hidden;
    height: auto;
    width: 100%;
}

.mytable {
    /* Consider removing float: left; and width: 100%; if not needed */
    font-family: Tahoma, Algerian, Geneva, sans-serif;
    width: 100%;
}

.mytable td {
    padding: 8px;  /* Adjusted padding for better spacing */
    border: 1px solid #dddfe1;
    border-left: none;
    border-right: none;
    height: 34px;  /* Set your desired height for the table cells */
    white-space: nowrap !important;  /* Prevent text from wrapping */
}

.mytable thead td {
    text-align: left;
    background-color: #386087;
    color: #ffffff;
    font-weight: bold;
    font-size: 13px;
    border: 1px solid #54585d;
}

.mytable tbody td {
    text-align: left !important;
    color: #636363;
}

.mytable tbody tr:nth-child(even) {
    background-color: white;
}

.mytable tbody tr:nth-child(even):hover {
    background-color: #e3f2fd;
}

.mytable tbody tr td:first-child {
    font-weight: bold;
}

"""
def create_container_html(container_name, container_keys, data, cfg):
    '''The first approach uses a generator expression inside the f-string, directly embedding the logic to create the table rows.'''
    hover_color = cfg.get('hover_color', '#e3f2fd')  # Default hover color
    
    #print(container_name)
    # Set the widths based on the container_name
    if container_name == 'container3a':
        stat_width = '25%'  # 1/4 width
        value_width = '75%'  # 3/4 width
    else:
        stat_width = '50%'  # Default width for other containers
        value_width = '50%'
    
    return f"""
        <div class="container container-200">
            <table class="mytable">
                {"".join(f"<tr><td style='width: {stat_width};'>{stat}</td><td style='width: {value_width}; text-align: center !important;'>{value}</td></tr>" for stat, value in data.items() if stat in container_keys)}
            </table>
        </div>
    """


def create_main_html(template_title, style, containers_html, pics, cfg):
    hover_color = cfg.get('hover_color', '#e3f2fd') #default
    img_width = "600px" if template_title == "val_tableC" else "500px"
    img_pad = "0px" if template_title == "val_tableC" else "0px"
    return f"""
        <!DOCTYPE html>
        <html>
        <head>
            <title>{template_title}</title>
            <style>{style}</style>
            <style>
                .mytable tbody tr:nth-child(even):hover {{
                    background-color: {hover_color};
                }}
                .mytable tbody tr:nth-child(odd):hover {{
                    background-color: {hover_color};
                }}
            </style>
        </head>
        <body class="my-custom-styles">
            <div class="flex-container" style="flex-wrap: wrap;">
                {containers_html}

                <div class="container-500">
                    <div class="img-container-cat" style="width:{img_width}; padding-left: {img_pad};"><img class="img-thumb-cat" style="" src="data:image/png;base64, {pics[0]}" alt="Histogram"></div>
                </div>
            </div>

            <div class="flex-container">
                <div class="container container-800">Hello</div>
                <div class="container">Con 2</div>
                <div class="container">Con 2</div>
                <div class="container">Con 2</div>
            </div>
            <!-- New container with bottom border -->
            <div class="container-gen" style="width: 100%; border-bottom: 1px solid #DDDFE1;"></div>
        </body>
        </html>
    """


def generate_table(data,container_keys,table_name, style=None, *pics, cfg=None):
    cfg = cfg or {}
    containers_html = "".join(create_container_html(container_name, container_keys, data, cfg) for container_name, container_keys in container_keys.items())
    return create_main_html(table_name, style, containers_html, pics, cfg)


##############################################################################################################################


def value_tableX(col,cfg,tokens=False):

    if tokens:
        dat = col
    else:
        dat = col.value_counts()

    # Separate the data into the first 20 rows and the rest
    top_20 = dat.head(10)
    other_values = dat.iloc[10:]

    # Create rows for the first 20 values
    rows_top_20 = [f"<tr><td>{idx[:20] if isinstance(idx, str) else idx}</td><td style='text-align: left !important;'>{val}</td><td>{valp}</td></tr>" for idx, val, valp in zip(top_20.index, top_20.values, np.round((top_20.values / col.size) * 100, 2))]

    # Create a row for the other values
    row_other_values = f"<tr><td>Other Values</td><td style='text-align: left !important;'>{other_values.sum()}</td><td>{np.round((other_values.sum() / col.size) * 100, 2)}</td></tr>"

    # Concatenate all rows
    all_rows = "".join(rows_top_20) + row_other_values

    top3 = col.head(3)
    low3 = col.tail(3)

    hover_color = cfg.get('hover_color','#e3f2fd')
    
    # Create a value table with f-strings (modify this as needed)
    value_table = f"""
        <!DOCTYPE html>
        <html>
        <head>
            <title>val_tableC</title>
            <style>
                .mycs body {{
                    font-family: Arial, sans-serif;
                }}

                .containerX {{
                    padding: 10px;
                    margin: 1px;
                    box-sizing: border-box;
                }}

                .flex-containerX {{
                    display: flex;
                    flex-wrap: wrap;
                }}

                .container-200X {{
                    flex: 2;
                }}

                .mytableX {{
                    float: left;
                    width: 100%;
                    font-family: Tahoma, Algerian, Geneva, sans-serif;
                }}

                .mytableX td {{
                    padding: 8px;
                    border: 1px solid #dddfe1;
                    border-left: none;
                    border-right: none;
                    height: 34px;
                    white-space: nowrap !important;
                }}

                .mytableX thead td {{
                    text-align: left !important;
                    background-color: #ffffff; /* White background color */
                    color: #386087; /* Your desired text color */
                    font-weight: bold;
                    font-size: 13px;
                    padding: 10px; /* Adjusted padding for better spacing */
                }}

                .mytableX tbody td {{
                    text-align: left !important;
                    color: #636363;
                }}

                .mytableX tbody tr:nth-child(even) {{
                    background-color: white;
                }}

                .mytableX tbody tr:nth-child(even):hover {{
                    background-color: {hover_color};
                }}

                .mytableX tbody tr:nth-child(odd):hover {{
                    background-color: {hover_color};
                }}

                .mytableX tbody tr td:first-child {{
                    font-weight: bold;
                }}
            </style>
        </head>
         <body class="mycs">
            <div class="flex-containerX">
                <div class="containerX container-200X">
                   <table class="mytableX">
                        <thead>
                            <tr>
                                <td>Values</td>
                                <td>Count</td>
                                <td>Frequency (%)</td>
                            </tr>
                        </thead>
                        {all_rows}
                    </table>
                </div>
            <div class="flex-containerX">
                <div class="containerX container-200X" style="width: 600px;">
                <div>
                <table class="mytableX" style="margin-bottom: 34px !important;">
                        <thead>
                            <tr>
                                <td style="width: 150px;">Head</td>
                                <td>Values</td>
                               
                            </tr>
                        </thead>
                        {''.join(f'<tr><td style="width: 150px;">{stat}</td><td style="text-align: left;">{value}</td></tr>' for stat, value in top3.items())}
                    </table>
                     </div>
                     <table class="mytableX">
                        <thead>
                            <tr>
                                <td style="width: 150px;">Tail</td>
                                <td>Values</td>
                                
                            </tr>
                        </thead>
                        {''.join(f'<tr><td style="width: 150px;">{stat}</td><td style="text-align: left;">{value}</td></tr>' for stat, value in low3.items())}
                    </table>
                     </div>
                    </div>
                     
            </div>
            <div class="container-x" style="width: 100%; border-bottom: 1px solid #DDDFE1;"></div>
            </div>
        </body>
        </html>
    """
    return value_table





from .dtype import *

import sys

class Theme:
    DEFAULT = {'mini_hist_color': '#c7e87c', 'tabset': '#c7e87c','hist_color':'#c7e87c','hist_kde_color':'yellow','bar_color':'#c7e87c','hover_color':'pink'}
    THEME1 = {'hist_color': 'pink', 'tabset': '#7CC7E8','hist_color':'#7CC7E8','hist_kde_color':'pink','bar_color':'#7CC7E8'}
    THEME2 = {'hist_color': 'blue', 'tabset': '#FF5733','hist_color':'#FF5733','hist_kde_color':'blue','bar_color':'#FF5733'}
    # Add more themes as needed
    

class Configs:

    # cache flag
    use_eda_cache = True  
    
    
    # Simple cache to store previously generated plots
    eda_cache = {}

    @classmethod
    def reset_cache(cls, df):
        # Iterate over keys in the cache and remove entries that correspond to the given DataFrame
        keys_to_remove = [key for key in cls.eda_cache.keys() if cls.eda_cache[key]['df'].equals(df)]
        for key in keys_to_remove:
            del cls.eda_cache[key]

        print(f"Cache reset for DataFrame with shape {df.shape}")





    ################################ setting DTYPES #####################################################

    
    dtypes_cache = {}
    dtype = None

    use_id = True
    
    @classmethod
    def get_dtypes(cls, df):
        """Returns a dict of suggested dtypes for each variable"""
        if cls.use_id:
            df_name = id(df)
        else:
            df_name = str(df)
        
        cls.dtype = cls.dtypes_cache.get(df_name,None)   
        
        if cls.dtype is not None:
            return cls.dtypes_cache[df_name]
        else:
            cls.dtypes_cache[df_name] = {i: determine_variable_type(df[i])[0] for i in df.columns}       
        return cls.dtypes_cache[df_name]


    @classmethod
    def set_dtypes(cls,df, dtypes_dict=None):
        """Updates the dtypes attribute with the provided dictionary"""
        if cls.use_id:
            df_name = id(df)
        else:
            df_name = str(df)

        if dtypes_dict == 'reset':
            cls.dtypes_cache[df_name] = None
               
        cls.dtype = cls.dtypes_cache.get(df_name,None)
        
        if cls.dtype is None or dtypes_dict is None:
            cls.dtype = cls.get_dtypes(df)  # Initialize dtypes using get_dtypes if it's None
        else:
            for key, new_value in dtypes_dict.items():
                if key in cls.dtype and cls.dtype[key] != new_value:
                    cls.dtype[key] = new_value
                    cls.dtypes_cache[df_name] = cls.dtype
                    print(cls.dtype)

        return 'done'



    
    ################################ THEME #####################################################
    
    current_theme = Theme.DEFAULT

    @classmethod
    def list_themes(cls, values=True):
        """
        List available themes and their values.

        Parameters:
        - values (bool, optional): If True, returns a dictionary with theme names and values.
          If False, returns a list of theme names only.

        Returns:
        - dict or list: If values is True, a dictionary of theme names and their values.
          If values is False, a list of theme names.
          
        """
        if values:
            themes = {}
            for attr in dir(Theme):
                if not callable(getattr(Theme, attr)) and not attr.startswith("__"):
                    themes[attr] = getattr(Theme, attr)
            return themes
        else:
            return [attr for attr in dir(Theme) if not callable(getattr(Theme, attr)) and not attr.startswith("__")]

    @classmethod
    def set_theme(cls, theme_name):
        """Set the current theme based on the provided theme name."""
        if hasattr(Theme, theme_name):
            cls.current_theme = getattr(Theme, theme_name)
        else:
            print(f"Theme '{theme_name}' not found. Using the default theme.")

    @classmethod
    def add_theme(cls, theme_name, theme_values):
        """
        Add a new theme to the Theme class.
    
        Parameters:
        - cls: The Theme class.
        - theme_name (str): The name of the new theme.
        - theme_values (dict): A dictionary containing the values of the new theme.
    
        Returns:
        - None
    
        Example:
        >>> Configs.add_theme('NEW_THEME', {'hist_color': 'purple', 'tabset': '#FFA500'})
        """
        setattr(Theme, theme_name, theme_values)


from ...dtype import *
from ...datahub import *



import pandas as pd
import numpy





def calc(df=None, x=None, y=None, cfg=None, dtypes=None, vars=None):
    """Call function based on the number of arguments (all, uni, bi)"""
    if df is not None and x is None and y is None:
        explore_univariate(df, dtypes,vars)
    elif x is not None and y is None:
        return calculate_univariate(df, x, cfg, dtypes)
    elif x is not None and y is not None:
        return calculate_bivariate(df, x, y, cfg, dtypes)
    else:
        raise ValueError("Invalid number of arguments")



def explore_univariate(df, dtypes, vars=None):
    """Explore univariate analysis for all columns"""
    from ..__init__ import eda
    if vars:
        dtypes = Configs.get_dtypes(df)
        selected_vars = [var for var, dtype in dtypes.items() if dtype == vars]
        
        print('dtypes',dtypes)
        print('sel.vars',selected_vars)

           
        for column in selected_vars:
            plot_instance = eda(df, column)
            plot_instance.show()
        
    else:
        for column in df.columns:
            plot_instance = eda(df, column)
            plot_instance.show()


def calculate_univariate(df, x, cfg, dtypes):
    """Calculate univariate analysis for a single variable"""
    col = df[x]
    dtype = get_dtype_info(col,x, dtypes)
    data = cont_comps(col, cfg, dtype)
    
    if dtype[0] == 'categorical' and data['MISSING'] > 0:
        col = col.dropna()
    
    return DataHub(col=col, data=data, variable_type=dtype[0])


def calculate_bivariate(df, x, y, cfg, dtypes):
    """Calculate bivariate analysis for two variables"""
    col1, dtype1 = get_dtype_info(df[x], dtypes)
    col2, dtype2 = get_dtype_info(df[y], dtypes)
    
    print(dtype1, dtype2)

    if dtype1[0] == 'continuous' and dtype2[0] == 'continuous':
        return DataHub(col1=col1, col2=col2, data=None, variable_type='biv_con')
    elif (dtype1[0] == 'continuous' and dtype2[0] == 'categorical') or (dtype2[0] == 'continuous' and dtype1[0] == 'categorical'):
        return DataHub(col1=col1, col2=col2, data=None, variable_type='biv_con_cat')
    elif dtype1[0] == 'categorical' and dtype2[0] == 'categorical':
        return DataHub(col1=col1, col2=col2, data=None, variable_type='biv_cat')
    else:
        print('compute_bivariate -> biv in progress')
        return None


def get_dtype_info(col,var_name, dtypes):
    """Get dtype information for a column"""
    if dtypes is None:
        return determine_variable_type(col)
    else:
        dtype_info = dtypes.get(var_name, determine_variable_type(col)[0])
        col_un = col.nunique()
        return dtype_info, col_un, col_un / len(col)



def cont_comps(ser = None, cfg = None, dtype = None):
    # ser = pd.series -> df.column => series
    
    def series_memory_usage(col, deep=False):
        total_memory_bytes = col.memory_usage(deep)

        if total_memory_bytes < 1_000_000:
            total_memory_mb = total_memory_bytes / (1024)
            return f"{total_memory_mb:.2f} KB"
        else:
            total_memory_mb = total_memory_bytes / (1024 ** 2)
            return f"{total_memory_mb:.2f} MB"

    dt = dtype[0]
    #print(dtype)
    if dt == 'unknown':
        return ser
    if isinstance(ser, np.ndarray):
        return ser
    else:
        data = {
            'TOTAL': ser.size,
            'VALUES': ser.count(),
            'VALUES_P': np.round((ser.count() / ser.size) * 100, 2),
            'MISSING': ser.isna().sum(),
            'MISSING_P': np.round((ser.isna().sum() / ser.size) * 100, 2),
            'DISTINCT': dtype[1],
            'DISTINCT_P': np.round(dtype[2] * 100, 2),
            '': '',
            'MEMORY': series_memory_usage(ser),
            'DTYPE': ser.dtype,
            'v_type': dt
        }
    
    
    if dt == 'categorical':
        #print('categorical is comming next')
        None
    elif dt == 'continuous':
        data.update({
            'MAX': np.round(np.nanmax(ser), 2),
            '95%': np.round(np.nanpercentile(ser, 95), 2),
            'Q3': np.round(np.nanpercentile(ser, 75), 2),
            'AVG': np.round(np.nanmean(ser), 1),
            'MEDIAN': np.round(np.nanmedian(ser), 1),
            'Q1': np.round(np.nanpercentile(ser, 25), 1),
            '5%': np.round(np.nanpercentile(ser, 5), 2),
            'MIN': np.round(np.nanmin(ser), 2),
            'RANGE': np.round(np.nanmax(ser) - np.nanmin(ser), 1),
            'IQR': np.round(np.nanpercentile(ser, 75) - np.nanpercentile(ser, 25), 1),
            'STD': np.round(np.nanstd(ser), 1),
            'VAR': np.round(np.nanvar(ser)),
            ' ': ' ',
            'KURT.': np.round(ser.kurtosis(), 3),
            'SKEW': np.round(ser.skew(), 3),
            'SUM': np.round(np.nansum(ser), 3)
        })

    elif dt == 'text': 
        w_len = ser.str.len()
        data.update({
        'w_len' : w_len,
        'Max length' : w_len.max(),
        'Mean length' : np.round(w_len.mean(),2),
        'Min length' : w_len.min()
        })

    elif dt == 'datetime':
        data.update({
            'MAX': ser.max(),
            # '95%': np.round(np.nanpercentile(ser, 95), 2),
            # 'Q3': np.round(np.nanpercentile(ser, 75), 2),
            # 'AVG': np.round(np.nanmean(ser), 1),
            # 'MEDIAN': np.round(np.nanmedian(ser), 1),
            # 'Q1': np.round(np.nanpercentile(ser, 25), 1),
            # '5%': np.round(np.nanpercentile(ser, 5), 2),
            'MIN': ser.min(),
            # 'RANGE': np.round(np.nanmax(ser) - np.nanmin(ser), 1),
            # 'IQR': np.round(np.nanpercentile(ser, 75) - np.nanpercentile(ser, 25), 1),
            # 'STD': np.round(np.nanstd(ser), 1),
            # 'VAR': np.round(np.nanvar(ser)),
            # ' ': ' ',
            # 'KURT.': np.round(ser.kurtosis(), 3),
            # 'SKEW': np.round(ser.skew(), 3),
            # 'SUM': np.round(np.nansum(ser), 3)
        })
    else:
        None
    
    return data

# data = np.random.randn(1000)  # Replace this with your dataset
# #compute(data,data).get('col')
# compute(data,data).get('data')
# compute(data,data).get('variable_type')




from pathlib import Path
import json
import os
import numpy as np
import pandas as pd


from .dtype import *


class DataHub(dict):
    """This class contains DataHub results."""

    def __init__(self, *args, **kwargs):
        if 'variable_type' in kwargs:
            
            super().__init__(**kwargs)
            self.variable_type = kwargs["variable_type"]
        else:
            raise ValueError("Unsupported initialization, missing visula type")
        

    def save(self, path = None):
        """
        Save DataHub to the specified path in JSON format.

        Parameters
        ----------
        path: Optional[str], default None
            The path where the DataHub will be saved.
        """
        if path is None:
            path = "DataHub.json"
        else:
            path = os.path.expanduser(path)

        with open(path, "w") as outfile:
            json.dump(self, outfile, cls=DataHubEncoder, indent=4)
        
        print(f"DataHub has been saved to {path}!")


import numpy as np
import pandas as pd



def determine_variable_type(data_series, threshold_unique=0.05, threshold_missing=0.2, threshold_text_ratio=0.1):
    """
    Determine if an independent variable is categorical, continuous, or text based on the data.

    Parameters:
    - data_series: A pandas Series representing the independent variable.
    - threshold_unique: A threshold for unique values (default is 0.05).
    - threshold_missing: A threshold for missing values (default is 0.2).
    - threshold_text_ratio: A threshold for the ratio of text values (default is 0.1).

    Returns:
    - Tuple containing the variable type ('categorical', 'continuous', 'text', 'datetime', or 'unknown') and the number of unique values.
    """
    
    if isinstance(data_series, np.ndarray):
        return 'unknown', None

    num_unique = data_series.nunique()
    unique_values_ratio = num_unique / len(data_series)

    if pd.api.types.is_numeric_dtype(data_series):
        missing_values_ratio = data_series.isnull().mean()

        if unique_values_ratio <= threshold_unique or missing_values_ratio >= threshold_missing:
            return 'categorical', num_unique, unique_values_ratio
        else:
            return 'continuous', num_unique, unique_values_ratio
            
    elif pd.api.types.is_object_dtype(data_series):    
        if unique_values_ratio >= threshold_text_ratio:
            return 'text', num_unique, unique_values_ratio
        else:
            return 'categorical', num_unique, unique_values_ratio
            
    elif pd.api.types.is_datetime64_any_dtype(data_series):
        return 'datetime', num_unique, unique_values_ratio

    else:
        return 'unknown', data_series  # Handle other data types if needed



# setup.py
from setuptools import setup, find_packages

setup(
    name="adix",
    version="0.0.0a1",
    author='imooger',
    description='Automatic data analysis tool',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
    ],
    install_requires=[
        # List any dependencies here
        'pandas>=2.1.0',
        'numpy>=1.26.2',
        'matplotlib>=3.8.2',
        'seaborn>=0.13.0',
        'scipy>=1.11.4',
        'Jinja2>=3.1.2',
        'notebook>=7.0.6',
        'wordcloud>=1.8.1',
    ],
    python_requires='>=3.10',
)